float water_getNoise(vec2 coord) {
    vec2 floored = floor(coord);
    vec4 samples = textureGather(noisetex, noiseResInverse * floored); // textureGather is slightly offset (at least on nvidia) and this offset can change with driver versions, which is why i floor the coords
    vec4 weights = (coord - floored).xxyy * vec4( 1,-1, 1,-1) + vec4(0, 1, 0, 1);
    weights *= weights * (-2.0 * weights + 3.0);
    return dot(samples, weights.yxxy * weights.zzww);
}

struct waveParams {
	vec2 inverseScale;
	vec2 scaledTranslation;
	vec2 skew;
	float height;
	bool sharpen;
	float sharpenThreshold;
	float sharpenMin;
};

float water_calculateWave(vec2 pos, const waveParams params) {
	pos = params.inverseScale * pos + params.scaledTranslation * frameTimeCounter;
	pos = pos.yx * params.skew + pos;
	float wave = water_getNoise(pos);
	if (params.sharpen)
		wave = 1.0 - almostIdentity(abs(wave * 2.0 - 1.0), params.sharpenThreshold, params.sharpenMin);
	return wave * params.height;
}

vec2 wavedx(vec2 position, vec2 direction, float speed, float frequency, float timeshift) {
	direction = normalize(direction);
    float x = dot(direction, position) * frequency + timeshift * speed;
    float wave = exp(sin(x) - 1.0);
    float dx = wave * cos(x);
    return vec2(wave, dx);
}

float getwaves(vec2 position, int iterations){
    const float dragMult = 0.058;
    position *= 0.1;
	float iter = 0.0;
    float phase = 6.0;
    float speed = 2.0 * Water_Speed;
    float weight = 1.0;
    float w = 0.0;
    float ws = 0.0;
    for(int i=0;i<iterations;i++){
        vec2 p = vec2(sin(iter), cos(iter)) * 30.0;
        vec2 res = wavedx(position, p, speed, phase, frameTimeCounter);
        position += normalize(position - p) * res.y * weight * dragMult;
        w += res.x * weight;
        iter += 12.0;
        ws += weight;
        weight = mix(weight, 0.0, 0.2);
        phase *= 1.18;
        speed *= 1.07;
    }
    return (w / ws) * 4.0;
}

float water_calculateWaves(vec2 pos) {
    
	const waveParams[4] params = waveParams[4](
		waveParams(1.0 / vec2(2.50, 3.33), vec2(0.40, 3.43) / vec2(2.50, 3.33), vec2(0.2, 1.3), 0.250,  true, 0.16, 0.08),
		waveParams(1.0 / vec2(0.71, 1.11), vec2(0.91,-0.71) / vec2(0.71, 1.11), vec2(0.0,-1.2), 0.120, false, 0.16, 0.08),
		waveParams(1.0 / vec2(0.26, 0.40), vec2(3.62, 0.26) / vec2(0.26, 0.40), vec2(0.0, 1.0), 0.020, false, 0.16, 0.08),
		waveParams(1.0 / vec2(0.09, 0.20), vec2(0.22, 0.16) / vec2(0.09, 0.20), vec2(0.0, 0.3), 0.008, false, 0.16, 0.08)
	);

	float waves = water_calculateWave(pos, params[0]) - params[0].height;
	for (int i = 1; i < params.length(); i++) {
		waves += water_calculateWave(pos, params[i]) - params[i].height;
	}
    
	return waves * 1.3;
}

vec3 calculateProceduralNormalsFlat(vec3 worldSpace) {
    vec2 coord = worldSpace.xz - worldSpace.y;

    float sampleC = water_calculateWaves(coord                 );
	float sampleX = water_calculateWaves(coord + vec2(0.2, 0.0));
	float sampleY = water_calculateWaves(coord + vec2(0.0, 0.7));

    vec3 waveNormals = normalize(vec3(
	    (sampleC - sampleX),
        (sampleC - sampleY),
        1.0
    ));

    return waveNormals;
}

#if !defined GBUFFERS_SHADOW && !defined COMPOSITE
vec3 water_calculateParallax(vec3 pos, vec3 direction) {
    const int steps = 5;

    vec3  interval = 0.4 * direction / abs(direction.y);
    vec3  offset   = vec3(0.0, 0.0, 0.0);
    float height   = water_calculateWaves(pos.xz);

    for (float i = 0.0; i < steps && height < offset.y; i++) {
        offset = (offset.y - height) * interval + offset;
        height = water_calculateWaves(offset.xz + pos.xz);
    }

    return vec3(1.0, 0.0, 1.0) * (offset) + pos;
}

void calculateProceduralNormals(inout vec3 normal) {
    if (materialIDs != 1) return;

    vec3 parallaxDirection   = tangentViewSpace;
    vec3 worldSpaceCorrected = worldSpace + gbufferModelViewInverse[3].xyz;
         worldSpaceCorrected = water_calculateParallax(worldSpaceCorrected, parallaxDirection.xzy);

    vec2 coord = worldSpaceCorrected.xz - worldSpaceCorrected.y;

    float sampleC = water_calculateWaves(coord                 );
	float sampleX = water_calculateWaves(coord + vec2(0.2, 0.0));
	float sampleY = water_calculateWaves(coord + vec2(0.0, 0.7));

    vec3 waveNormals = normalize(vec3(
	    (sampleC - sampleX),
        (sampleC - sampleY),
        1.0
    ));

    normal = waveNormals;
}
#endif