vec4 calculateSunSpot(vec3 viewVector, vec3 sunVector) {
	float cosTheta = dot(viewVector, sunVector);

	// limb darkening approximation
	const vec3 a = vec3(0.397, 0.503, 0.652);
	const vec3 halfa = a * 0.5;
	const vec3 normalizationConst = vec3(0.896, 0.873, 0.844); // probably not set correctly

	float x = clamp(1.0 - (sunAngularRadius - acos(cosTheta)) / sunAngularRadius, 0.0, 1.0);
	vec3 sunDisk = exp2(log2(-x * x + 1.0) * halfa) / normalizationConst;

	//return cosTheta > cos(sunAngularRadius) ? sunDisk : background;
    return vec4(sunDisk, float(cosTheta > cos(sunAngularRadius)));
}

vec3 sky_atmosphere(
    vec3 background, 
    vec3 viewVector, 
    vec3 sunVector, vec3 moonVector, vec3 earthPosition,
    const ivec2 steps, 
    const vec2 skyDepths
) {
    // Setup scattering / extinction integral
	const mat2x3 a = mat2x3(-rayleighCoefficient * RCPLOG2, -mieExtinction * RCPLOG2);
	const mat2x3 b = mat2x3(-1.0 / rayleighCoefficient, -1.0 / mieExtinction);
	const mat2x3 c = mat2x3( 1.0 / rayleighCoefficient,  1.0 / mieExtinction);

    float VoL    = dot(viewVector, sunVector);
    vec2  phaseS = sky_phase(min1(( VoL + sinSunAngle) / (1.0 + sinSunAngle)), mieG);
    vec2  phaseM = sky_phase(min1((-VoL + sinSunAngle) / (1.0 + sinSunAngle)), mieG);

    float stepSize     = (skyDepths.y - skyDepths.x) / steps.x;
    vec3  rayIncrement = viewVector * stepSize;
    vec3  rayPosition  = viewVector * skyDepths.x + earthPosition + rayIncrement * 0.0;

    vec3  scatterSun     = vec3(0.0, 0.0, 0.0);
    vec3  scatterMoon    = vec3(0.0, 0.0, 0.0);
    vec3  scatterAmbient = vec3(0.0, 0.0, 0.0);
    vec3  inTransmission = vec3(1.0, 1.0, 1.0);


    for (int x = 0; x < steps.x; x++, rayPosition += rayIncrement) {
        float rayLengthSq  = dot(rayPosition, rayPosition);
        float rayLength    = sqrt(rayLengthSq);
        float inAltitude   = rayLength - earthRadius;

        vec3  inOpticalDensities   = calculateOpticalDensities(inAltitude);
        if (inOpticalDensities.y > 1e35) {
            inTransmission *= 0.0;
            break;
        }
              inOpticalDensities  *= stepSize;
        vec3  outOpticalDensitiesS = vec3(0.0, 0.0, 0.0);
        vec3  outOpticalDensitiesM = vec3(0.0, 0.0, 0.0);

		vec3 stepOpticalDepth = extinctionCoefficients * inOpticalDensities;

		vec3 stepTransmittance       = exp(-stepOpticalDepth);
		vec3 stepTransmittedFraction = (stepTransmittance - 1.0) / -stepOpticalDepth;
		vec3 stepScatteringVisible   = inTransmission * stepTransmittedFraction;

        vec2   outStepLength   = dot(rayPosition, sunVector) * vec2(1.0, -1.0);
               outStepLength   = sqrt(outStepLength * outStepLength + atmosphereRadiusSq - rayLengthSq) - outStepLength;
        vec2   outStepSize     = outStepLength / steps.y;
        mat2x3 outRayIncrement = mat2x3(sunVector * outStepSize.x, moonVector * outStepSize.y);
        mat2x3 outRayPosition  = mat2x3(rayPosition + outRayIncrement[0] * 0.5, rayPosition + outRayIncrement[1] * 0.5);

        for (int y = 0; y < steps.y; y++, outRayPosition += outRayIncrement) {
            vec2 outRayLength = vec2(length(outRayPosition[0]), length(outRayPosition[1]));
            vec2 outAltitude  = outRayLength - earthRadius;

            outOpticalDensitiesS += calculateOpticalDensities(outAltitude.x);
            outOpticalDensitiesM += calculateOpticalDensities(outAltitude.y);
        }

        outOpticalDensitiesS *= outStepSize.x;
        outOpticalDensitiesM *= outStepSize.y;

        mat2x3 scatteringScale = mat2x3(
            scatterCoefficients[0] * (exp2(a[0] * inOpticalDensities.x) * b[0] + c[0]),
            scatterCoefficients[1] * (exp2(a[1] * inOpticalDensities.y) * b[1] + c[1])
        );

        vec3 outTransmissionS = exp(extinctionCoefficients * -outOpticalDensitiesS);
        vec3 outTransmissionM = exp(extinctionCoefficients * -outOpticalDensitiesM);
        
        scatterSun     += scatteringScale * phaseS * stepScatteringVisible * outTransmissionS;
        scatterMoon    += scatteringScale * phaseM * stepScatteringVisible * outTransmissionM;
        scatterAmbient += scatteringScale * vec2(0.25) * stepScatteringVisible;
        inTransmission *= stepTransmittance;
    }

    vec3 atmosphere = background * inTransmission + mix(scatterSun * sunTint + scatterMoon * moonTint + scatterAmbient * lightColors[3], lightColors[3] * 0.25 * RCPPI, rainStrength);


    return atmosphere;
}

void applyAtmosphere(in param sky, inout vec3 color, in vec3 viewVector, in vec3 sunVector, in vec2 depths, in ivec2 quality, in float skyMask, in float lightMask) {
    if (skyMask > 0.5) return;

    // Calculate Backdrop
    vec4 sunSpot  = calculateSunSpot(viewVector,  sunVector) * vec4(sunTint  * 250.0, lightMask);
    vec4 moonSpot = calculateSunSpot(viewVector, -sunVector) * vec4(moonTint * 100.0, lightMask);

    color = mix(color, sunSpot.rgb,  sunSpot.a);
    color = mix(color, moonSpot.rgb, moonSpot.a);
    color = sky_atmosphere(color, viewVector, sunVector, -sunVector, sky.earthPosition, quality, depths);
}