const float atmosphereHeight   = 110e3;
const float earthRadius        = 6731e3;
const float atmosphereRadius   = earthRadius + atmosphereHeight;
const float atmosphereRadiusSq = atmosphereRadius * atmosphereRadius;

const vec3 scaleHeights    = vec3(8.0e3, 1.2e3, 1111.0);
const vec3 scaleHeightsRCP = 1.0 / scaleHeights;

const float airNumberDensity       = 2.6867774e19; // Couldn't find it for air, so just using value for an ideal gas. Not sure how different it is for actual air.
const float ozoneConcentrationPeak = 16e-6;
const float ozoneNumberDensity     = airNumberDensity * ozoneConcentrationPeak;
const vec3  ozoneCrossSection      = vec3(4.51103766177301E-21, 3.2854797958699E-21, 1.96774621921165E-22);
const vec3  ozoneAbsorption        = ozoneCrossSection * ozoneNumberDensity;

const vec3 rayleighCoefficient = vec3(5.8e-6, 13.5e-6, 33.1e-6); //vec3(4.593e-6, 1.097e-5, 2.716e-5); //vec3(5.8e-6, 13.5e-6, 33.1e-6);

const vec3 mieCoefficient = mix(vec3(2.500e-6, 2.500e-6, 2.500e-6) * 0.1, vec3(4.500e-5, 4.500e-5, 4.500e-5), rainStrength);
const vec3 mieExtinction  = mieCoefficient * 1.11;

float mieG = 0.395;

const mat2x3 scatterCoefficients    = mat2x3(rayleighCoefficient, mieCoefficient);
const mat3x3 extinctionCoefficients = mat3x3(rayleighCoefficient, mieExtinction, ozoneAbsorption);


struct param {
    vec3 earthPosition;
    bool sphereIntersect;
    vec2 sphereDepths;        // X = Entry depth || Y = Exit depth
};

param calculateSkyParameters(vec3 viewVector, vec3 upVec) {
    param sky;
    sky.earthPosition   = upVec * (earthRadius + cameraPosition.y + horizonAltitude);
    sky.sphereIntersect = raySphereIntersect(viewVector, sky.earthPosition, atmosphereRadius, sky.sphereDepths);
    sky.sphereIntersect = sky.sphereIntersect.x && bool(greaterThan(sky.sphereDepths, vec2(0.0)));
    sky.sphereDepths    = max0(sky.sphereDepths);

    return sky;
}

float rayleighPhase(float cosTheta) {
    return (3.0 / (16.0 * PI))  * (1.0 + cosTheta * cosTheta);
}

float hgPhase(float cosTheta, const float g) {
	const float gg = g * g;
	const float rGG = 0.2 / gg;
	const float p1 = (5.375 * (1.10 - gg)) * RCPPI * 0.2 * rGG;
	float p2 = (cosTheta * cosTheta + 1.0) * pow(-2.0 * g * cosTheta + 1.0 + gg, -1.5);
	return p1 * p2;
}

vec2 sky_phase(float cosTheta, const float g) {
    vec2 phase   = vec2(rayleighPhase(cosTheta), hgPhase(cosTheta, g));
         phase.y = mix(phase.y, 0.25, rainStrength);
	return phase;
}

vec3 calculateOpticalDensities(float altitude) {
    // https://ozonewatch.gsfc.nasa.gov/facts/images/ozone_concentration_graph.gif
    // https://www.desmos.com/calculator/lbc39zsxwg
    vec3 densities   = exp(-altitude * scaleHeightsRCP);
         densities.z = exp(-max(35e3 - altitude, 0.0) / 5000.0) * exp(-max(altitude - 35e3, 0.0) / 15000.0);
    return densities;
}

vec3 transmittedScatteringIntegral(float od, const vec3 coeff) {
	const vec3 a = -coeff / log(2.0);
	const vec3 b = -1.0 / coeff;
	const vec3 c =  1.0 / coeff;

	return exp2(a * od) * b + c; // = ∫{0,od} e^(-coeff*x) dx
}