/* DRAWBUFFERS:45 */

layout (location = 0) out vec4 colortex4Write;
layout (location = 1) out vec4 colortex5Write;


const bool colortex4MipmapEnabled = true;


in vec2 texcoord;

flat in mat4 shadowViewMatrix;
flat in mat4x3 lightColors;
flat in mat2x3 sunVec;
flat in mat2x3 lightVec;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex7;
uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D noisetex;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferPreviousProjection, gbufferPreviousProjectionInverse;
uniform mat4 gbufferPreviousModelView, gbufferPreviousModelViewInverse;
uniform mat4 shadowProjection, shadowProjectionInverse;
uniform vec3 cameraPosition, previousCameraPosition;
uniform vec3 upPosition;
uniform float rainStrength;
uniform float viewWidth, viewHeight;
uniform float frameTimeCounter;
uniform int frameCounter;

uniform ivec2 eyeBrightnessSmooth;

const vec2 pixel = 1.0 / vec2(viewWidth, viewHeight);


#include "/lib/Global.glsl"

#include "/lib/Utility/Space_Transformations.glsl"
#include "/lib/Utility/Raytrace.glsl"
#include "/lib/Utility/Intersect.glsl"

#include "/lib/Shared/Atmosphere_Params.glsl"

#include "/lib/Fragment/Mapping.glsl"
#include "/lib/Fragment/Sampling_Patterns.glsl"
#include "/lib/Fragment/Shadows.glsl"
#include "/lib/Fragment/BSDF.glsl"
#include "/lib/Fragment/3D_Clouds.glsl"
#include "/lib/Fragment/Volumetric_Light.glsl"


float calculateAO(gbuffer info, material mat, scene spaces, vec2 pattern) {
    if (mat.emitter > 0.5) return 1.0;

    #ifndef Ambient_Occlusion 
    return 1.0;
    #endif

    const float ditherSize = pow2(16.0);
    float dither = ditherSize * pattern.x;

    float FOV    = degrees(atan(1.0 / gbufferProjection[1].y) * 4.0);
    vec3  normal = mat3(gbufferModelView) * info.normal[0];

    const float radius   = 1.5;
    const float strength = AO_Strength * 50.0;

    #ifdef Stochastic_AO 
    radius = 1.1;
    strength = 3.0;
    #endif

    float dx = radius / spaces.view[1].z * 90.0 / FOV;

    float AO = 0.0;

    for (int i = 0; i < AO_Quality; i++) {
        vec2 dir = fermatsSpiralGoldenS(i * ditherSize + dither, AO_Quality * ditherSize) * dx;
        vec2 h   = texcoord + dir;

        vec3 nvpos = calculateViewSpace(vec3(h, texture2D(depthtex1, h).x)).xyz;

        float d    = length(nvpos - spaces.view[1]);

        AO += max(0.0, dot(normal, nvpos - spaces.view[1]) / d - 0.15)
            * max(0.0, - (d - radius))
            * float(d > 0.00001);
    }

    return pow(1.0 - clamp01(AO / AO_Quality), strength);
}

vec3 calculateGlobalIllumination(gbuffer info, material mat, scene spaces, vec2 pattern) {
    #ifndef RSM
    return vec3(0.0);
    #endif

    vec3 shadowSpace = calculateShadowSpace(spaces.world[1]);

    if (calculateShadow(shadowtex0, distortShadowSpace(shadowSpace)) * max0(dot(lightVec[1], info.normal[0])) >= 1.0) return vec3(0.0);

    const float ditherSize = pow2(64.0);
    float dither = ditherSize * pattern.y + 0.5;

    vec3  shadowNormal = normalize(mat3(shadowViewMatrix) * info.normal[0]);

    vec3  light  = vec3(0.0);
    float weight = 1.0;
    float lod    = 0.0;
    
    float radius = RSM_Radius * shadowMapResInverse;

    if (eyeBrightnessSmooth.y <= 100) radius = radius * 0.2;

    for (int i = 0; i < RSM_Rays; i++) {
        vec2 offset = fermatsSpiralGoldenS(i * ditherSize + dither, RSM_Rays * ditherSize) * radius;
        if (dot(shadowNormal.xy, offset) <= 0.0) offset = -offset;

        vec3 offsetPosition = shadowSpace + vec3(offset, 0.0);
        vec3 offsetPositionDistorted = distortShadowSpace(offsetPosition);

        float shadowDepth = texture2DLod(shadowtex0, offsetPositionDistorted.xy, lod).x;
              shadowDepth = (((shadowDepth * 2.0 - 1.0) * Shadow_Z_Stretch) * 0.5 + 0.5);

        vec3  samplePosition = vec3(offsetPosition.xy, shadowDepth) - shadowSpace;
        float sampleLength   = length(samplePosition);

        vec3 samplePositionNormalized = normalize(samplePosition) * vec3(1.0, 1.0, -1.0);

        if (sampleLength > radius) continue;
        weight++;

        float falloff = 1.0 / (1.0 + pow2(sampleLength) * shadowMapResolution);

        float diffuse = max0(dot(samplePositionNormalized, shadowNormal));

        if (diffuse <= 0.0) continue;

        vec3 normalSample = texture2DLod(shadowcolor1, offsetPositionDistorted.xy, lod).rgb * 2.0 - 1.0;
             normalSample = -normalSample;

        float diffuse_bounce = max0(dot(samplePositionNormalized, normalSample));

        if (diffuse_bounce <= 0.0) continue;

        vec3 lightSample = toLinear(texture2DLod(shadowcolor0, offsetPositionDistorted.xy, lod).rgb) * falloff * diffuse * diffuse_bounce;
        if (lightSample != lightSample) lightSample = vec3(1.0);

        light += lightSample * RSM_Brightness * 3;
    }

    return light / weight;
}

vec4 accumulateLighting(gbuffer info, material mat, scene spaces, vec2 pattern) {
    if (mat.land < 0.5) return vec4(0.0, 0.0, 0.0, 1.0);
    const vec2 scaleTransformAdjust = -vec2(0.0, pixel.y * texcoord.y * C0_Scale);
    vec2  texcoordPrevoius  = -spaces.cameraVelocity + texcoord;
    vec4  previousFrameInfo = texture2DLod(colortex4, texcoordPrevoius * C0_Scale + scaleTransformAdjust, 0.0);

    //float AO        = calculateAO(info, mat, spaces, pattern);
    vec3  lighting  = calculateGlobalIllumination(info, mat, spaces, pattern);

    if (lighting != lighting) lighting = vec3(0.0, 0.0, 0.0);

    float weight = floor(texcoordPrevoius) == 0 ? Accumulator_Weight : 0.0;
    
    vec3  accumulated = mix(lighting, previousFrameInfo.rgb, weight);

    return vec4(accumulated, 1.0);
}

vec4 packVolumetrics(gbuffer info, material mat, scene spaces, vec2 pattern) {
    vec4 vc = calculateVolumetricClouds(info, mat, spaces, pattern);
    vec4 vl = calculateVolumetricLight(info, mat, spaces, pattern);

    bool  depthOrder  = cameraPosition.y < Volumetric_Clouds_Altitude;
    float compression = length(lightColors[0] + lightColors[1]);

    vec3  scatter      = depthOrder ? vc.rgb * vl.a + vl.rgb : vl.rgb * vc.a + vc.rgb;
    float transmission = vc.a * vl.a;

    return vec4(sqrt(scatter / compression), transmission); // Normalize, compress gamma
}


void main() {
    mat2 coord = mat2(
        texcoord,
        texcoord
    );

    gbuffer  info   = calculateGbufferInformation(coord);
    material mat    = calculateMaterialMasks(info.specular, info.color, info.depthtex, info.materialIDs, info.gbufferForward && !info.forwardSolidMask);
    scene    spaces = calculateSpaces(coord, info.depthtex);

    vec2 pattern = calculateAlteredSamplingDistributors();

    colortex4Write = accumulateLighting(info, mat, spaces, pattern);
    colortex5Write = packVolumetrics(info, mat, spaces, pattern);
}