/* DRAWBUFFERS:01 */

layout (location = 0) out vec4 colortex0Write;
layout (location = 1) out vec4 colortex1Write;


const bool colortex4MipmapEnabled = true;

const bool shadowtex0Mipmap    = true;
const bool shadowtex1Mipmap    = true;
const bool shadowcolor0Mipmap  = true;
const bool shadowcolor1Mipmap  = true;


in vec2 texcoord;

flat in mat4 shadowViewMatrix;
flat in mat4x3 lightColors;
flat in mat2x3 sunVec;
flat in mat2x3 lightVec;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D shadowcolor1;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D noisetex;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;
uniform mat4 shadowProjection, shadowProjectionInverse;
uniform vec3 cameraPosition, previousCameraPosition;
uniform float viewWidth, viewHeight;
uniform float rainStrength;
uniform float frameTimeCounter;
uniform int frameCounter;

const vec2 pixel = 1.0 / vec2(viewWidth, viewHeight);


#include "/lib/Global.glsl"

#include "/lib/Utility/Space_Transformations.glsl"
#include "/lib/Utility/Intersect.glsl"

#include "/lib/Shared/Atmosphere_Params.glsl"
#include "/lib/Shared/Atmosphere_Compute.glsl"

#include "/lib/Fragment/Mapping.glsl"
#include "/lib/Fragment/Sampling_Patterns.glsl"
#include "/lib/Fragment/Shadows.glsl"
#include "/lib/Fragment/BSDF.glsl"
#include "/lib/Fragment/3D_Clouds.glsl"


void setupSolidAlbedo(inout gbuffer info) {
         if (info.forwardSolidMask) info.color[0] = info.color[1];
    else if (info.depthtex.y >= 1.0) info.color[0] = vec4(0.0, 0.0, 0.0, 1.0);
}

float calculateBlockerDepth(vec3 shadowSpace, const float ditherSize, float dither) {
    #ifndef PCSS
        return 0.3;
    #endif

    const float maxBlockerDist = 48.0;
    const int   searchSamples  = 7;

    float searchScale  = 0.5 * maxBlockerDist / abs(shadowProjectionInverse[2].z);
    float spread       = tan(sunAngularRadius) * abs(shadowProjectionInverse[2].z) * shadowProjection[0].x;
    float searchRadius = spread * searchScale;

    float lod = log2(2.0 * shadowMapResolution * searchRadius / sqrt(searchSamples));

    float blockerDepth = 0.0;

    for (int i = 0; i < searchSamples; i++) {
        vec2 offset = fermatsSpiralGoldenN(i * ditherSize + dither, searchSamples * ditherSize) * searchRadius;
        vec3 shadowSpaceDistorted = distortShadowSpace(shadowSpace + vec3(offset, 0.0));

        float shadowDepth = shadowSpaceDistorted.z - texture2DLod(shadowtex1, shadowSpaceDistorted.xy, lod).x;

        blockerDepth += shadowDepth * Shadow_Z_Stretch;
    }

    return min(blockerDepth / searchSamples, searchScale) * spread * shadowMapResolution;
}

vec3 calculateSunlightVisibility(gbuffer info, material mat, scene spaces, vec3 shadowSpace, vec2 pattern) {
    const float ditherSize = pow2(64.0);
    float dither = ditherSize * pattern.y;

    float NdotL = dot(info.normal[0], lightVec[1]);
    if (NdotL < 0.0 && mat.translucent < 0.5) return vec3(0.0);

    int   shadowQuality        = 5;
    float inverseShadowQuality = 1.0 / float(shadowQuality);
    float filterSize           = calculateBlockerDepth(shadowSpace, ditherSize, dither);

    vec3  shadow     = vec3(0.0);

    float shadowBias = 0.0002 / Shadow_Z_Stretch;

    for (int i = 0; i < shadowQuality; i++) {
        vec2 offset = fermatsSpiralGoldenN(i * ditherSize + dither, shadowQuality * ditherSize) * filterSize * shadowMapResInverse;

        vec3  shadowSpaceDistorted = distortShadowSpace(shadowSpace + vec3(offset, 0.0));
        vec2  zStep = shadowSpaceDistorted.z - vec2(shadowBias, shadowBias * 2.0);

        if (shadowSpaceDistorted.z >= 1.0) {
            shadow = vec3(shadowQuality);
            break;
        }

        float shadowDepth0 = texture2DLod(shadowtex0, shadowSpaceDistorted.xy, 0.0).x;
        float shadowDepth1 = texture2DLod(shadowtex1, shadowSpaceDistorted.xy, 0.0).x;
        vec4  shadowColor  = texture2DLod(shadowcolor0, floor(shadowSpaceDistorted.xy * shadowMapResolution) * shadowMapResInverse, 0.0);

        float shadow0 = step(zStep.x, shadowDepth0);
        float shadow1 = step(zStep.x, shadowDepth1);

        float transparent = (1.0 - step(zStep.y, shadowDepth0)) * step(zStep.y, shadowDepth1);
        float waterDepth  = (shadowDepth1 - shadowDepth0) * Shadow_Z_Stretch;
              waterDepth  = (waterDepth * shadowProjectionInverse[2].z + shadowProjection[3].z);
        
        vec3 shadowSample = shadow0 + (1.0 - shadow0) * shadow1 * shadowColor.rgb;

        if (transparent * (1.0 - floor(shadowColor.a)) > 0.1) {
            shadowSample *= exp(waterDepth * (waterAbsorbCoeff + waterScatterCoeff));
        }

        shadow += shadowSample;
    }

    shadow *= inverseShadowQuality;

    #ifdef Parallax_Occlusion
        shadow.rgb *= info.shadowed;
    #endif

    #ifdef HRTS
        const float HRTS_Render_Distance = HRTS_Range;

        vec3 rayOrigin    = spaces.view[1];
        vec3 rayDirection = -lightVec[0];

        if (bool(greaterThan(shadow.rgb, vec3(0.0))) && length(rayOrigin) < HRTS_Render_Distance && mat.translucent < 0.5) {
            const int   HRTS_steps   = HRTS_Samples;
            const float R_HRTS_steps = 1.0 / HRTS_steps;

            float traceDist = -0.1;

            vec3 rayIncrement = rayDirection * R_HRTS_steps * traceDist;
            vec3 rayPosition  = rayOrigin + rayIncrement * pattern.y;

            bool hit = false;

            for (int i = 0; i < HRTS_steps; i++) {
                vec3  rayPositionScreen = calculateScreenSpace(rayPosition);

                float rayDepth   = linearizeDepth(texture2D(depthtex1, rayPositionScreen.xy).x);
                float difference = rayDepth - rayPosition.z;

                if (difference > 0.01 && difference < -traceDist ) { hit = true; break; }

                rayPosition += rayIncrement;
            }

            shadow.rgb *= float(!hit);
        }
    #endif

    return shadow.rgb;
    //return vec3(filterSize);
}

vec4 denoise(sampler2D colortex) {
    vec4 accumulated = texture2DLod(colortex, (texcoord - temporalJitter()) * C0_Scale, 0.0);
    return accumulated;
}

void calculateShading(inout gbuffer info, material mat, scene spaces, vec2 pattern, inout float shadows) {
    if (mat.land < 0.5) return;

    vec3  V   = normalize(-spaces.world[2]);

    float fresnel = f_dielectric(spaces.world[2], info.normal[1], mat.F0.y) * mat.reflectiveness.y;

    float cloudShadow    = calculateCloudShadows(spaces.world[1] + cameraPosition);
    vec3  shadowSpace    = calculateShadowSpace(spaces.world[1]);
    vec3  shadowLightmap = calculateSunlightVisibility(info, mat, spaces, shadowSpace, pattern) * cloudShadow;
    vec4  bakedLighting  = denoise(colortex4);

    vec3 emission      = Emissive_Mapping == Resource_Pack ? info.color[0].rgb * 3 * Emissive_Brightness : torchColor * 2;
         emission     *= mat.emissiveness.x;

    float normalOcclusion = mix(pow(min(-info.normal[0].y * 0.5 + 5.0, 1.0), 1.0), 1.0, 0.0);
    vec3  fakeBounce      = vec3(0.05, 0.045, 0.03) * lightColors[2] * (-info.normal[0].y * 0.7 + 0.9) * 0.6 * pow2(bakedLighting.a);

    float diffuseUp   = mix(sqrt( info.normal[0].y * 0.5 + 0.5), RCPPI, 0.5);
    float diffuseDown = mix(sqrt(-info.normal[0].y * 0.5 + 0.5), RCPPI, 0.5);

    vec3 totalShading  = vec3(0.0);
         totalShading += diffuse(mat, info.color[0].rgb, V, lightVec[1], info.normal[0], lightColors[2], shadowLightmap) * PI;
         totalShading += bakedLighting.rgb * cloudShadow * lightColors[2] * PI;
         totalShading += torchColor * info.lightmap[0].x * bakedLighting.a;
         #ifdef Skylight
         totalShading += lightColors[3] * diffuseUp   * pow2(info.lightmap[0].y) * bakedLighting.a * RCPPI * Skylight_Brightness;
         totalShading += lightColors[2] * diffuseDown * info.lightmap[0].y * bakedLighting.a * RCPPI * 0.25  * Skylight_Brightness;
         //totalShading += (lightColors[3] * normalOcclusion + fakeBounce) * info.lightmap[0].y * bakedLighting.a * RCPPI * Skylight_Brightness;
         #endif
         totalShading *= max0(1.0 - mat.emissiveness.x);
         totalShading *= 1.0 - fresnel;

    if (totalShading != totalShading) totalShading = vec3(0.0, 0.0, 0.0);
    
    info.color[0].rgb = info.color[0].rgb * totalShading + emission;
    //info.color[0].rgb = vec3(bakedLighting.a);

    if (info.gbufferForward) {
        shadows = calculateShadow(shadowtex0, distortShadowSpace(calculateShadowSpace(spaces.world[0]))) * calculateCloudShadows(spaces.world[0] + cameraPosition);
    } else {
        shadows = length(shadowLightmap);
    }
}


void main() {
    mat2 coord = mat2(
        texcoord,
        texcoord
    );

    gbuffer  info   = calculateGbufferInformation(coord);
    if (info.depthtex.y >= 1.0) {
        colortex0Write = vec4(0.0);
        colortex1Write = texture2D(colortex1, coord[0]);
        return;
    }
    material mat    = calculateMaterialMasks(info.specular, info.color, info.depthtex, info.materialIDs, info.gbufferForward && !info.forwardSolidMask);
    scene    spaces = calculateSpaces(coord, info.depthtex);

    vec2 pattern = calculateJitteredSamplingDistributors();

    setupSolidAlbedo(info);
    
    float shadows = 0.0;
    calculateShading(info, mat, spaces, pattern, shadows);

    colortex0Write = info.color[0];
    colortex1Write = vec4(texture2D(colortex1, coord[0]).rgb, shadows * 0.5 + 0.5);
}