/* DRAWBUFFERS:0 */

layout (location = 0) out vec4 colortex0Write;


const bool colortex0MipmapEnabled = true;
const bool colortex4MipmapEnabled = true;


in vec2 texcoord;

flat in mat4 shadowViewMatrix;
flat in mat4x3 lightColors;
flat in mat2x3 sunVec;
flat in mat2x3 lightVec;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex4;
uniform sampler2D colortex5;
uniform sampler2D shadowtex0;
uniform sampler2D shadowtex1;
uniform sampler2D shadowcolor0;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D noisetex;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;
uniform mat4 shadowProjection, shadowProjectionInverse;
uniform vec3 cameraPosition, previousCameraPosition;
uniform vec3 upPosition;
uniform ivec2 eyeBrightnessSmooth;
uniform float rainStrength;
uniform float viewWidth, viewHeight;
uniform float frameTimeCounter;
uniform int frameCounter;
uniform int isEyeInWater;

const vec2 pixel = 1.0 / vec2(viewWidth, viewHeight);


#include "/lib/Global.glsl"

#include "/lib/Utility/Space_Transformations.glsl"
#include "/lib/Utility/Raytrace.glsl"
#include "/lib/Utility/Intersect.glsl"

#include "/lib/Shared/Atmosphere_Params.glsl"
#include "/lib/Shared/Atmosphere_Compute.glsl"

#include "/lib/Fragment/Sampling_Patterns.glsl"
#include "/lib/Fragment/Shadows.glsl"
#include "/lib/Fragment/2D_Clouds.glsl"
#include "/lib/Fragment/3D_Clouds.glsl"
#include "/lib/Fragment/BSDF.glsl"
#include "/lib/Fragment/Mapping.glsl"


vec2 calculateRefraction(vec2 coord) {
    vec2 depthtex = vec2(
        texture2D(depthtex0, coord).x,
        texture2D(depthtex1, coord).x
    );

    if (depthtex.x > depthtex.y || texture2D(colortex3, coord).a >= 1.0) return coord;

    mat2x3 viewSpace  = mat2x3(
        calculateViewSpace(vec3(coord, depthtex.x)),
        calculateViewSpace(vec3(coord, depthtex.y))
    );

    vec3 normalBump = decodeNormal3x16(texture2D(colortex2, coord).x);
    vec3 normalFlat = mat3(gbufferModelViewInverse) * normalize(cross(dFdx(viewSpace[0]), dFdy(viewSpace[0])));

    float refractionDepth = min(distance(viewSpace[0], viewSpace[1]), 10.0);

    vec2  refractedCoord = coord + (normalBump.xz - normalFlat.xz) * refractionDepth / viewSpace[0].z * 0.06;
    float refractedDepth = texture2D(depthtex1, refractedCoord).x;

    return depthtex.x > refractedDepth ? coord : refractedCoord;
}

void generateStars(inout vec3 color, in vec3 worldVector, in float visibility) {
    if (visibility >= 1.0) return;

    const float res = 360.0;

    vec3 p  = worldVector * res;
    vec3 id = floor(p);
    vec3 fp = fract(p) - 0.5;

    float rp    = hash13(id);
    float stars = 1.0 - smoothstep(0.3, 0.5, length(fp));

    color = vec3(stars) * step(rp, 0.003) * 0.1 * Stars_Brightness * 113;
}

void calculateWaterShading(inout gbuffer info, material mat, scene spaces) {
    if (mat.water < 0.5 || isEyeInWater > 0.5) return;

    vec3 totalCoeff   = waterAbsorbCoeff + waterScatterCoeff * vec3(0.0, 0.3, 1.0);

    float waterDepth  = distance(spaces.world[0], spaces.world[1]) * 1.5;

    vec3 sunLight     = info.lightmap[1].y * lightColors[2] * 0.35;
    vec3 torchLight   = info.lightmap[1].x * torchColor;
    vec3 skyLight     = info.lightmap[1].y * lightColors[3];

    vec3 finalShading = sunLight + torchLight + skyLight;

    vec3 absorption   = exp(-totalCoeff * waterDepth);
    vec3 scatter      = transmittedScatteringIntegral(waterDepth, totalCoeff) / PI * LOG2 * finalShading * waterScatterCoeff * 0.5;

    info.color[0].rgb = info.color[0].rgb * absorption + scatter;
    info.color[0].rgb = mix(info.color[0].rgb, scatter, 0.2);
}

void drawTransparents(inout gbuffer info, material mat, scene spaces) {
    if (!info.gbufferForward) return;

    vec3  extinctionCoeff  = (1.0 / powf(info.color[1].rgb, 0.25)) * 1.0;
    float transparentDepth = min1(distance(spaces.world[0], spaces.world[1]));

    info.color[0].rgb *= exp(-transparentDepth * extinctionCoeff);
}

void calculateSpecularLight(inout gbuffer info, material mat, scene spaces, param sky, vec2 pattern) {
    if (info.depthtex.x >= 1.0) return;
    const ivec2 atmosphereQuality = ivec2(4, 2);
    float compression = length(lightColors[0] + lightColors[1]);

    bool  importantPixels = mat.reflectiveness.y > 0.37; //exp(-1) = 0.3678

    vec3  normal    = mat3(gbufferModelView) * info.normal[1];
    vec3  up        = normalize(upPosition);
    vec3  rayOrigin = spaces.screen[0];

    vec3  H   = normalize(lightVec[0] - spaces.view[2]);
    float NoH = min1(dot(normal, H));
    float NoL = dot(lightVec[0], normal);
    float NoV = dot(-spaces.view[2], normal);

    vec3  reflectedSunlight = BRDF_GGX(NoH, NoL, NoV, mat.roughness.y) * lightColors[2];
    
    vec3 reflection = vec3(0.0);

    int rays = importantPixels ? Specular_Rays : 1;

    for (int i = 0; i < rays; i++) {
        vec3 ray;
        vec3 facetNormal = normal;
        bool intersect   = false;

        vec3 reflectionSample = vec3(0.0);

        if (importantPixels) {
                 facetNormal  = distributeMicrofacets(normal, hash42(vec2(i + pattern.y, frameCounter)), mat.roughness.y);
            vec3 rayDirection = reflect(spaces.view[2], facetNormal);

            intersect = raytraceIntersection(rayOrigin, rayDirection, ray, 0.5, 6.0, 4.0);
            if (intersect) {
                float LOD = calculateGGXMipLevel(spaces.view[1], calculateViewSpace(ray), mat.roughness.y);
                reflectionSample = texture2DLod(colortex0, ray.xy, LOD).rgb;
            } else if (isEyeInWater < 0.5) {
                param reflectedSky = calculateSkyParameters(rayDirection, up);
                applyAtmosphere(sky, reflectionSample, rayDirection, sunVec[0], reflectedSky.sphereDepths, atmosphereQuality, 0.0, 0.0);

                if (floor(ray.xy) == vec2(0.0)) {
                    vec4 volumetricSample     = texture2DLod(colortex5, ray.xy * C0_Scale, 0.0);
                         volumetricSample.rgb = pow2(volumetricSample.rgb) * compression;
                         reflectionSample     = reflectionSample * volumetricSample.a + volumetricSample.rgb;
                }

                reflectionSample *= pow2(info.lightmap[1].y);
            }
        }
        reflectionSample += reflectedSunlight * info.shadowed;
        reflection       += reflectionSample * f_dielectric(spaces.view[2], normal, mat.F0.y) * mat.reflectiveness.y;
    }

    reflection /= rays;

    if (reflection != reflection) reflection = vec3(0.0, 0.0, 0.0);

    info.color[0].rgb += reflection / (1.0 + PI * mat.roughness.y);
}

void calculateUnderwaterShading(inout gbuffer info, material mat, scene spaces) {
    if (isEyeInWater < 0.5) return;

    vec3 totalCoeff   = waterAbsorbCoeff + waterScatterCoeff * vec3(0.0, 0.3, 1.0);

    float waterDepth  = length(spaces.world[0]) * 1.5;

    vec3 finalShading = lightColors[3];

    vec3 absorption   = exp(-totalCoeff * waterDepth);
    vec3 scatter      = transmittedScatteringIntegral(waterDepth, totalCoeff) / PI * LOG2 * finalShading * waterScatterCoeff * 0.5;

    info.color[0].rgb = info.color[0].rgb * absorption + scatter;
    info.color[0].rgb = mix(info.color[0].rgb, scatter, 0.2);
}

void unpackVolumetrics(inout gbuffer info) {
    const float RCP_C0_Scale = 1.0 / C0_Scale;
    const vec2 scaledPixel = pixel * RCP_C0_Scale;
    vec2  scaledCoord = texcoord * C0_Scale;
    vec2  sdCoord     = floor(gl_FragCoord.xy * C0_Scale) * RCP_C0_Scale * pixel;
    float compression = length(lightColors[0] + lightColors[1]);

    float depth = linearizeDepth(info.depthtex.x);

    vec4  colorSample     = texture2D(colortex5, scaledCoord);
          colorSample.rgb = pow2(colorSample.rgb) * compression;
    vec4  denoised        = colorSample;
    float weight          = 1.0;

    for (int x = -1; x <= 1; x++) {
        for (int y = -1; y <= 1; y++) {
            vec2 offset     = vec2(x, y) * scaledPixel;
            vec2 coord      = offset + scaledCoord;
            vec4 lookup     = texture2D(colortex5, coord);
                 lookup.rgb = pow2(lookup.rgb) * compression;
            
            float sampleDepth = linearizeDepth(texture2D(depthtex0, sdCoord + offset).x);
            if (sampleDepth < depth - 0.1 || sampleDepth > depth + 0.1)
                continue;

            denoised.rgb  = min(denoised.rgb, lookup.rgb);
            denoised.a   += lookup.a;
            weight ++;
        }
    }

    denoised.a /= weight;

    colorSample = mix(colorSample, denoised, 0.5);

    info.color[0].rgb = info.color[0].rgb * colorSample.a + colorSample.rgb;
}


void main() {
    mat2 coord = mat2(
        texcoord,
        calculateRefraction(texcoord)
    );

    gbuffer  info   = calculateGbufferInformation(coord);
    material mat    = calculateMaterialMasks(info.specular, info.color, info.depthtex, info.materialIDs, info.gbufferForward && !info.forwardSolidMask);
    scene    spaces = calculateSpaces(coord, info.depthtex);
    param    sky    = calculateSkyParameters(spaces.view[2], normalize(upPosition));

    vec2 pattern = calculateJitteredSamplingDistributors();


    generateStars(info.color[0].rgb, spaces.world[2], mat.land);
    applyAtmosphere(sky, info.color[0].rgb, spaces.view[2], sunVec[0], sky.sphereDepths, ivec2((Atmosphere_Steps), 2), mat.land, 1.0);
    calculateCirrusClouds(info.color[0].rgb, spaces.world[1], spaces.world[2], mat.land);
    calculateWaterShading(info, mat, spaces);
    drawTransparents(info, mat, spaces);
    calculateSpecularLight(info, mat, spaces, sky, pattern);
    calculateUnderwaterShading(info, mat, spaces);
    unpackVolumetrics(info);


    if (info.color[0] != info.color[0]) info.color[0] = vec4(0.0, 0.0, 0.0, 1.0);
    colortex0Write = info.color[0];
}