/* DRAWBUFFERS:06 */

layout (location = 0) out vec4 colortex0Write;
layout (location = 1) out vec4 colortex6Write;


const bool colortex0MipmapEnabled = true;
const bool colortex6MipmapEnabled = true;


in vec2 texcoord;

uniform sampler2D colortex0;
uniform sampler2D colortex1;
uniform sampler2D colortex2;
uniform sampler2D colortex3;
uniform sampler2D colortex6;
uniform sampler2D depthtex0;
uniform sampler2D depthtex1;
uniform sampler2D noisetex;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;
uniform vec3 cameraPosition, previousCameraPosition;
uniform float viewWidth, viewHeight;
uniform int frameCounter;

const vec2 pixel = 1.0 / vec2(viewWidth, viewHeight);


#include "/lib/Global.glsl"

#include "/lib/Utility/Space_Transformations.glsl"

#include "/lib/Fragment/Sampling_Patterns.glsl"


vec3 calculateTAA(gbuffer info, scene spaces, vec2 pattern) {
    vec3 screenPos = vec3(texcoord, info.depthtex.y);
    vec2 prevCoord = -spaces.cameraVelocity + texcoord;
    vec2 blurDir   = vec2(0.0);

    #ifdef Motion_Blur
        blurDir = -spaces.cameraVelocity * pattern.x;
    #endif

    float weight = floor(prevCoord) == 0 ? dot(0.5 - abs(fract(prevCoord / pixel) - 0.5), vec2(1.0)) : 0.0;

    mat2x3 limits = mat2x3(0.0);

    for (int i = -1; i <= 1; ++i) {
        for (int j = -1; j <= 1; ++j) {
            if (ivec2(i, j) == ivec2(0, 0)) continue;

            vec3 currentSample = texture2DLod(colortex0, ivec2(i, j) * pixel + texcoord + blurDir, 0.0).rgb;

            if (ivec2(i, j) == ivec2(-1, -1)) {
                limits = mat2x3(currentSample, currentSample);

                continue;
            }

            limits[0] = min(currentSample, limits[0]);
            limits[1] = max(currentSample, limits[1]);
        }
    }

    vec3 antiAliased = mix(texture2DLod(colortex0, texcoord + blurDir, 0.0).rgb, clamp(texture2DLod(colortex6, prevCoord, 0.0).rgb, limits[0], limits[1]), sqrt(weight) * 0.5 + inversesqrt(weight * 2.0 + 4.0));
    
    return antiAliased;
}

vec2 calculateBlurTileOffset(const int id) {
	// offsets approximately follows a multiple of this: (1 - 4⁻ˣ) / 3

	const vec2 idMult = floor(id * 0.5 + vec2(0.0, 0.5));
	const vec2 offset = vec2(1.0, 2.0) * (1.0 - exp2(-2.0 * idMult)) / 3.0;

	const vec2 paddingPixels = vec2(2.0, 9.0); // Set as needed
	const vec2 paddingAccum  = idMult * paddingPixels;

	return offset + paddingAccum * pixel;
}

vec3 calculateBloomTile(vec2 coord, const float lod) {
	coord *= exp2(lod);

	if (clamp01(coord) != coord) return vec3(0.0);

    vec2 resolution = pixel * exp2(lod);

    vec3  bloom       = vec3(0.0);
	float totalWeight = 0.0;
	
	for (int y = -3; y <= 3; y++) {
		for (int x = -3; x <= 3; x++) {
			float weight  = clamp01(1.0 - length(vec2(x, y)) / 4.0);
			      weight *= weight;
			
			bloom += texture2DLod(colortex6, coord + vec2(x, y) * resolution, lod).rgb * weight;
			totalWeight += weight;
		}
	}
	
	return bloom / totalWeight * Bloom_Brightness * 0.3;
}

vec3 calculateFlareTile(vec2 coord, const float lod, const float power, float exposure) {
	coord *= exp2(lod);

	if (clamp01(coord) != coord) return vec3(0.0);

    vec2 resolution = pixel * exp2(lod);

    vec3  bloom       = vec3(0.0);
	float totalWeight = 0.0;
	
	for (int y = -3; y <= 3; y++) {
		for (int x = -3; x <= 3; x++) {
			float weight  = pow2(clamp01(1.0 - length(vec2(x, y)) / 4.0));
			
			bloom += powf(texture2DLod(colortex6, coord + vec2(x, y) * resolution, lod).rgb / exposure, power) * weight;
			totalWeight += weight;
		}
	}
	
	return bloom / totalWeight;
}

vec3 calculateBloomTiles() {
    vec3 blurTiles = calculateBloomTile(texcoord - calculateBlurTileOffset(0), 1);
		for (int i = 1; i < 6; blurTiles += calculateBloomTile(texcoord - calculateBlurTileOffset(i), ++i));
    
    vec2 flareTileOffset = vec2(0.5 + pixel.x * 4.0, 0.0);
    vec3 flareTile = calculateFlareTile(texcoord - flareTileOffset, 4, 2.0, texture2DLod(colortex6, vec2(0.0), 0.0).a);

    return blurTiles + flareTile;
}

float calculateAverageLuminance() {
    const float maxLumaRange  = 5.0;
    const float minLumaRange  = 0.01;
    const float exposureDecay = 0.01;

    float avglod = int(exp2(min(viewWidth, viewHeight))) - 1;

	float averagePrevious = texture2DLod(colortex6, vec2(0.0), 0.0).a;
    float averageCurrent  = clamp(sqrt(dot(texture2DLod(colortex0, vec2(0.5), avglod).rgb, lumaCoeff)), minLumaRange, maxLumaRange);

    return mix(averagePrevious, averageCurrent, exposureDecay);
}


void main() {
    mat2 coord = mat2(
        texcoord,
        texcoord
    );

    gbuffer  info   = calculateGbufferInformation(coord);
    scene    spaces = calculateSpaces(coord, info.depthtex);

    vec2 pattern = calculateAlteredSamplingDistributors();

    vec3 color = calculateTAA(info, spaces, pattern);

    colortex0Write = vec4(calculateBloomTiles(), 1.0);
    colortex6Write = vec4(color, calculateAverageLuminance());
}