/* DRAWBUFFERS:0 */

layout (location = 0) out vec4 outColor;


in vec2 texcoord;

uniform sampler2D colortex0;
uniform sampler2D colortex6;
uniform sampler2D colortex7;
uniform sampler2D noisetex;
uniform sampler2D depthtex1;

uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform mat4 gbufferPreviousModelView, gbufferPreviousProjection;
uniform float frameTimeCounter;
uniform float centerDepthSmooth;
uniform float aspectRatio;
uniform float viewWidth, viewHeight;
uniform float near, far;
uniform int frameCounter;

const vec2 res   = vec2(viewWidth, viewHeight);
const vec2 pixel = 1.0 / res;


#include "/lib/Utility/Space_Transformations.glsl"

#include "/lib/Shared/ACES_Main.glsl"

#include "/lib/Texturing/Smoothing.glsl"

#include "/lib/Fragment/Mapping.glsl"
#include "/lib/Fragment/Sampling_Patterns.glsl"


float ld(float depth) {
    return (2.0 * near) / (far + near - depth * (far - near));
}

float tubeBorder() {
	vec2 a = texcoord * (1.0 - texcoord) * (35.0 + texcoord);
    return mix(2.0 - pow(a.x * a.y, 0.25), 1.0, 0.9);
}

vec2 calculateDistortion() {
    vec2 coord = texcoord * 2.0 - 1.0;
    #ifdef Helmet_Distortion
	    coord *= tubeBorder();
    #endif

    return coord * 0.5 + 0.5;
}

void calculateDepthOfField(inout vec3 color, in vec2 coord, in vec2 pattern) {
    #ifndef DepthOfField
        return;
    #endif

    const float apertureRadiusMM = DOF_Aperture * 0.5;
    const float sensorHeightMM   = 24;
    const float apertureRadius   = apertureRadiusMM / 1000.0;
    const float sensorHeight     = sensorHeightMM / 1000.0;

    float focalLength  = 0.5 * sensorHeight * gbufferProjection[1].y;
    float expDepth     = texture2D(depthtex1, coord).x;
    float depth        = linearizeDepth(expDepth);
    float focus        = linearizeDepth(centerDepthSmooth);

    float CoC          = apertureRadius * focalLength * (depth - focus) / (depth * (focus - focalLength) * sensorHeight); // COC radius in units of the sensor height
    vec2  CoCAspect    = CoC * vec2(1.0 / aspectRatio, 1.0) * vec2(1.0 / DOF_Anamorphic, DOF_Anamorphic);

    #ifdef Distance_Blur
          CoC         += smoothstep(0.1, 1.8, ld(expDepth)) * 1.4;
    #endif

    #ifdef Tilt_Shift
          CoC          = coord.y * 2.0 - 1.0;
    #endif
    
    const float error      = 1.1;
    const float errorRCP   = 1.0 / error;
    const float ditherSize = pow2(64.0);
    const float total      = DepthOfFieldQuality * ditherSize;
    float dither           = ditherSize * pattern.y + 0.5;
    vec3  depthOfField     = vec3(0.0);

    for (int i = 0; i < DepthOfFieldQuality; i++) {
        float index   = i * ditherSize + dither;
	    float theta   = index * GOLDEN_ANGLE;
        vec2  offset  = sincos(theta) * (error - pow(index / total, DOF_Bias)) * errorRCP * CoCAspect;
        depthOfField += texture2DLod(colortex6, offset + coord, 0.0).rgb;
    }

    color = depthOfField / DepthOfFieldQuality;
}

vec3 dirtyLens(vec3 color) {
    #ifndef Dirty_Lens
        return color;
    #endif
    
    const float res = 1.0 / 4.0;

    vec2 p = gl_FragCoord.xy * res;
    vec2 id = floor(p);
    vec2 fp = fract(p) - 0.5;

    float rp = hash12(id);
    float stars = 1.0 - smoothstep(0.0, 0.5, length(fp));

    return color * (1.0 + stars * step(rp, 0.003) * 1);
}

vec2 calculateBlurTileOffset(const int id) {
	// offsets approximately follows a multiple of this: (1 - 4⁻ˣ) / 3

	const vec2 idMult = floor(id * 0.5 + vec2(0.0, 0.5));
	const vec2 offset = vec2(1.0, 2.0) * (1.0 - exp2(-2.0 * idMult)) / 3.0;

	const vec2 paddingPixels = vec2(2.0, 9.0); // Set as needed
	const vec2 paddingAccum  = idMult * paddingPixels;

	return offset + paddingAccum * pixel;
}

void calculateBloom(inout vec3 color, in vec2 coord) {
    vec3 bloom = bicubicTexture(colortex0, coord * exp2(-1.0) + calculateBlurTileOffset(0)).rgb;
    float totalWeight = 1.0;
    for (int i = 1; i < 6; ++i) {
        float tileWeight = 1.0 / (i + 1.0);
        bloom += bicubicTexture(colortex0, coord * exp2(-i - 1.0) + calculateBlurTileOffset(i)).rgb * tileWeight;
        totalWeight += tileWeight;
    }
    bloom /= totalWeight;

    color = mix(color, bloom, 0.1 * Bloom_Brightness);
}

void calculateDiffractionSpikes(inout vec3 color, in vec2 coord, in vec2 pattern) {
    #ifndef Diffraction_Spikes
        return;
    #endif

    const int   tileID     = 1;
    const vec2  tileOffset = calculateBlurTileOffset(tileID);
    const float tileScale  = exp2(-tileID - 1.0);

    const float qualityRCP = 1.0 / Diffraction_Spikes_Quality;
    const float falloff    = 2.0;

    const float rotStep = radians(360.0 / Aperture_Blades);
    float rotDir = rotStep + radians(Aperture_Rotation);

    float radius   = 0.3 * qualityRCP;
    vec2  radScale = radius * vec2(1.0 / aspectRatio, 1.0);

    float totalWeight = 1.0;
    vec3  spikes = vec3(0.0);

    for (int x = 0; x < Aperture_Blades; x++, rotDir += rotStep) {
        vec2 offset = sincos(rotDir) * radScale * pattern.x;
        vec2 coord  = texcoord + offset;
        for (int y = 0; y < Diffraction_Spikes_Quality; y++, coord += offset) {
            float weight = pow(1.0 - y * qualityRCP, falloff);
            bool  onTile = floor(coord) == vec2(0.0);

            if (!onTile) continue;

            spikes      += texture2DLod(colortex0, coord * tileScale + tileOffset, 0.0).rgb * weight;
            totalWeight += weight;
        }
    }

    color += spikes / totalWeight * 0.1;
}

vec2 calculateFlareCoord(vec2 coord, float scale, float distortion, out float onTile) {
    vec2 flareCoord = coord * 2.0 - 1.0;

    // remap ghost position
    flareCoord = flareCoord * scale;
    
    // barrel distortion
    vec2  ratio = vec2(1.0, 1.0 / aspectRatio) * vec2(1.0 / DOF_Anamorphic, DOF_Anamorphic);
    float dist  = dot(flareCoord * ratio, flareCoord * ratio);
    flareCoord  = flareCoord * pow(dist, distortion);
    
    vec2 xyDist = abs(flareCoord);

         flareCoord = flareCoord * 0.5 + 0.5;

    onTile  = smoothstep(1.0, 0.85, xyDist.x) * smoothstep(1.0, 0.85, xyDist.y);

    return flareCoord * exp2(-4.0) + vec2(0.5 + pixel.x * 4.0, 0.0);
}

void calculateLensFlare(inout vec3 color, in vec2 coord) {
    #ifndef Lens_Flare
        return;
    #endif
    
    float onTile;
    vec2  flareCoord;
    vec3  flare = vec3(0.0);

    flareCoord = calculateFlareCoord(coord, -1.00, 3.0, onTile);
    flare     += bicubicTexture(colortex0, flareCoord).rgb * onTile * vec3(0.3, 0.0, 0.0);

    flareCoord = calculateFlareCoord(coord, -1.02, 3.0, onTile);
    flare     += bicubicTexture(colortex0, flareCoord).rgb * onTile * vec3(0.0, 0.3, 0.0);

    flareCoord = calculateFlareCoord(coord, -1.04, 3.0, onTile);
    flare     += bicubicTexture(colortex0, flareCoord).rgb * onTile * vec3(0.0, 0.0, 0.3);

    flareCoord = calculateFlareCoord(coord,  7.0, 0.5, onTile);
    flare     += bicubicTexture(colortex0, flareCoord).rgb * onTile * vec3(1.0, 0.5, 0.2);

    flareCoord = calculateFlareCoord(coord, -2.0, 0.4, onTile);
    flare     += bicubicTexture(colortex0, flareCoord).rgb * onTile * vec3(1.0, 0.2, 0.2);

    flareCoord = calculateFlareCoord(coord, -4.0, 0.4, onTile);
    flare     += bicubicTexture(colortex0, flareCoord).rgb * onTile * vec3(0.2, 1.0, 0.3);

    color += flare * 0.0001;
}

void calculateExposure(inout vec3 color) {
    color = (color / texture2D(colortex6, vec2(0.0)).a) * Exposure;
}

void calculateBorder(inout vec3 color) {
    #ifndef Helmet_Distortion
        return;
    #endif

    color *= smoothstep(1.0, 0.9, tubeBorder());
}

void tonemap_filmic(inout vec3 color) {
	const vec3 a = vec3(1.00, 1.00, 1.00);
	const vec3 b = vec3(1.00, 1.00, 1.00);
	const vec3 c = vec3(1.00, 1.00, 1.00);
	const vec3 d = vec3(1.00, 1.00, 1.00);

	vec3 cr = mix(vec3(dot(color, lumaCoeff)), color, d) + 1.0;

	color = pow(color, a);
	color = pow(color / (1.0 + color), b / a);
	color = pow(color * color * (-2.0 * color + 3.0), cr / c);
}

void lookup(inout vec3 color) {
    const vec2 inverseSize = vec2(1.0 / 512, 1.0 / 4096);
    const mat2 correctGrid = mat2(
        vec2(1.0, inverseSize.y * 512),
        vec2(0.0, Selected_LUT * inverseSize.y * 512)
    );

    #ifndef LUT
        return;
    #elif Show_LUT
        if (gl_FragCoord.x < 512.0 && gl_FragCoord.y < 512.0) {
            vec2 uv = gl_FragCoord.xy / 512.0;
            color = texture2D(colortex7, uv + correctGrid[1]).rgb;
            return;
        }
    #endif

    color = clamp01(color);

    float blueColor = color.b * 63.0;

    vec4 quad = vec4(0.0);
    quad.y = floor(floor(blueColor) * 0.125);
    quad.x = floor(blueColor) - (quad.y * 8.0);
    quad.w = floor(ceil(blueColor) * 0.125);
    quad.z = ceil(blueColor) - (quad.w * 8.0);

    vec4 texPos = (quad * 0.125) + (0.123046875 * color.rg).xyxy + 0.0009765625;

    vec3 newColor1 = texture2D(colortex7, texPos.xy * correctGrid[0] + correctGrid[1]).rgb;
    vec3 newColor2 = texture2D(colortex7, texPos.zw * correctGrid[0] + correctGrid[1]).rgb;
    
    color = mix(newColor1, newColor2, fract(blueColor));
}

void ditherScreen(inout vec3 color) {
    vec3 lestynRGB = vec3(dot(vec2(171.0, 231.0), gl_FragCoord.xy));
         lestynRGB = fract(lestynRGB.rgb / vec3(103.0, 71.0, 97.0));

    color += lestynRGB.rgb / 255.0;
}

void calculateNightEye(inout vec3 color) {
    float luminance  = dot(color, lumaCoeff);
    float lows       = exp(-luminance * 40.0);
    vec3  saturation = mix(vec3(1.0, 1.0, 1.0), vec3(0.0, 0.0, 0.0), lows);
    vec3  tint       = normalize(vec3(0.2, 0.7, 1.0));

    color = mix(luminance * tint, color, saturation);
}


void main() {
    vec2 pattern           = calculateSamplingDistributors();
    vec2 distortedTexcoord = calculateDistortion();

    vec3 color = texture2D(colortex6, distortedTexcoord).rgb;

    calculateDepthOfField(color, distortedTexcoord, pattern);
    calculateBloom(color, distortedTexcoord);
    calculateDiffractionSpikes(color, distortedTexcoord, pattern);
    calculateLensFlare(color, distortedTexcoord);
    calculateNightEye(color);
    calculateBorder(color);
    calculateExposure(color);

    #ifdef ACES_Tonemap
    ColorCorrection m;
	m.lum = vec3(0.2125, 0.7154, 0.0721);
	m.saturation = 1.0 + SAT_MOD;
	m.vibrance = VIB_MOD;
	m.contrast = 1.0 - CONT_MOD;
	m.contrastMidpoint = CONT_MIDPOINT;

	m.gain = vec3(1.1, 1.0, 1.0) + GAIN_MOD; //Tint Adjustment
	m.lift = vec3(0.0, 0.0, 0.0) + LIFT_MOD * 0.1; //Tint Adjustment
	m.InvGamma = vec3(1.0, 1.0, 1.0);

    FilmToneMap(color);

    color = WhiteBalance(color);
	color = Vibrance(color, m);
	color = Saturation(color, m);
    color = Contrast(color, m);
    color = LiftGammaGain(color, m);
    #endif

    #ifdef Filmic_Tonemap
    tonemap_filmic(color);
    #endif

	color = toSRGB(color);
    lookup(color);
    ditherScreen(color);

    outColor = vec4(color, 1.0);
}