/* DRAWBUFFERS:01 */

layout (location = 0) out vec4 colortex0Write;
layout (location = 1) out vec4 colortex1Write;


in vec4 color;
in vec3 worldSpace;
in vec3 viewSpace;
in vec3 tangentViewSpace;
in vec2 texcoord;
in vec2 vertexLightmaps;

flat in mat3 tbnMatrixWorld;
flat in mat3 tbnMatrixView;
flat in mat3x2 atlasTileInfo;
flat in vec2 midCoord;
flat in vec2 tileSize;
flat in float parallaxDepth;
flat in int materialIDs;

uniform sampler2D texture;
uniform sampler2D normals;
uniform sampler2D specular;
uniform sampler2D noisetex;
uniform mat4 gbufferModelView, gbufferModelViewInverse;
uniform mat4 gbufferProjection, gbufferProjectionInverse;
uniform vec3 shadowLightPosition;
uniform float viewWidth, viewHeight;
uniform int frameCounter;

const vec2 pixel = 1.0 / vec2(viewWidth, viewHeight);

#define atlasTileOffset     atlasTileInfo[0]
#define atlasTileSize       atlasTileInfo[1]
#define atlasTileResolution atlasTileInfo[2]


#include "/lib/Utility/Temporal.glsl"

#include "/lib/Fragment/Sampling_Patterns.glsl"


float parallax_calculateShadow(vec3 coord, mat2 tcd, vec3 dir) {
    if (materialIDs == 2) return 1.0;
    if (dir.z <= 0.0) return 0.0;

    coord.xy = (coord.xy - atlasTileOffset) / atlasTileSize;

    vec3 increment = vec3(Parallax_Depth, Parallax_Depth, 1.0) * (dir / dir.z) / 256;
    vec3 offset = vec3(0.0, 0.0, coord.z);

    for (int i = 0; i < 64 && offset.z < 1.0; i++) {
        offset += increment;

        float foundHeight = textureGrad(normals, fract(coord.xy + offset.xy) * atlasTileSize + atlasTileOffset, tcd[0], tcd[1]).a;
        if (offset.z < foundHeight - 4.0/256.0) return 0.0;
    }

    return 1.0;
}

vec2 parallax_calculateCoordinate(vec2 textureCoordinates, mat2 textureCoordinateDerivatives, vec3 tangentViewVector, inout float parallaxShadow) {
    #if !defined Parallax_Occlusion || !defined GBUFFERS_TERRAIN
        return textureCoordinates;
    #endif

    if (tangentViewVector.z > 0.0) return textureCoordinates;

    ivec2 textureResolution = textureSize2D(texture, 0);

    mat2 derivatives = textureCoordinateDerivatives; derivatives[0] *= textureResolution; derivatives[1] *= textureResolution;
    float d = max(max(dot(derivatives[0], derivatives[0]), dot(derivatives[1], derivatives[1])), 1.0);
    vec2 atlasTileResolutionLod = atlasTileResolution * exp2(ceil(-0.5 * log2(d)));

    tangentViewVector.xy *= Parallax_Depth;
    tangentViewVector.y  *= atlasTileResolution.x / atlasTileResolution.y; // Tile aspect ratio - fixes snow layer sides
    vec2 texelDelta = tangentViewVector.xy * atlasTileResolutionLod;

    vec3 position = vec3((textureCoordinates - atlasTileOffset) / atlasTileSize, 1.0);
    float height;

    for (int i = 0; i < 1024; ++i) {
        height = texture2DGrad(normals, fract(position.st) * atlasTileSize + atlasTileOffset, textureCoordinateDerivatives[0], textureCoordinateDerivatives[1]).a;

        // Calculate distance to the next two texels
        // xy = closest, zw = second closest
        vec4 texelCoordinates = fract(position.stst * atlasTileResolutionLod.stst);
        texelCoordinates = texelCoordinates * vec4(2.0, 2.0, 1.0, 1.0) + vec4(-1.0, -1.0, -0.5, -0.5);
        texelCoordinates = (texelCoordinates * sign(texelDelta).xyxy) * -0.5 + 0.5;

        vec4 distanceToNextTexelDirectional = texelCoordinates / abs(texelDelta).xyxy;
        float distanceToNextTexel = min(distanceToNextTexelDirectional.x, distanceToNextTexelDirectional.y);

        // Check for intersection
        if (tangentViewVector.z * distanceToNextTexel + position.z < height) break;

        // Distance to second closest texel
        float distanceToNextTexel2 = min(
            maxof(distanceToNextTexelDirectional.xy),
            minof(distanceToNextTexelDirectional.zw)
        );

        // Move to halfway trough next texel
        position += 0.5 * (distanceToNextTexel + distanceToNextTexel2) * tangentViewVector;
    }

    position.xy = clamp(fract(position.st), 1e-6, 1.0 - 1e-6) * atlasTileSize + atlasTileOffset;

    parallaxShadow = parallax_calculateShadow(position, textureCoordinateDerivatives, (shadowLightPosition) * tbnMatrixView);

    return position.xy;
}

vec4 getAlbedo(vec2 coord, mat2 derivatives, vec2 pattern) {
    vec4 albedo     = texture2DGrad(texture, coord, derivatives[0], derivatives[1]);
         albedo.rgb = toLinear(albedo.rgb * color.rgb);
    return clamp01(albedo);
}

vec4 getTangentNormals(vec2 coord) {
    vec4 normal = texture2D(normals,  coord) * 2.0 - 1.0;
    normal.xyz  = clampNormal(normal.xyz, -normalize(tangentViewSpace));
    return normal;
}

vec3 calculateSpecularMaps(vec2 coord) {
    vec3 specular = texture2D(specular, coord).rgb;
    #ifdef Gray_Scale_Specular_Support
         specular = vec3(specular.r, 0.0, 0.0);
    #endif
    return specular;
}

vec2 calculateLightmaps(vec3 normalTex, vec3 normal, vec2 pattern) {
    vec2 lightmaps   = pow(vertexLightmaps, vec2(1.0, 3.0));
         lightmaps.x = 16.0 - min(15.0, (lightmaps.x - 0.5 / 16.0) * 16.0 * 16.0 / 15.0);
	     lightmaps.x = pow2(clamp01(1.0 - pow4(lightmaps.x / 16.0))) / (1.0 + pow2(lightmaps.x));
         lightmaps.x = mix(lightmaps.x, 0.2, float(materialIDs == 3));

    return sqrt(lightmaps * pow(vec2(color.a), vec2(1.0, 1.5))) + pattern.x / vec2(31,63);
}


void main() {
    float shadowed    = 1.0;
    mat2  derivatives = mat2(dFdx(texcoord), dFdy(texcoord));
    vec2  coord       = parallax_calculateCoordinate(texcoord, derivatives, tangentViewSpace, shadowed);

    vec2 pattern   = calculateJitteredSamplingDistributors();

    vec4 albedo    = getAlbedo(coord, derivatives, pattern);
    vec4 normalTex = getTangentNormals(coord);
    vec3 specular  = calculateSpecularMaps(coord);
    vec3 normal    = normalize(tbnMatrixWorld * normalTex.xyz);
    vec2 lightmaps = calculateLightmaps(normalTex.xyz, normal, pattern);


    colortex0Write   = albedo;
    colortex1Write   = vec4(
        encodeNormal3x16(normal),
        encode3x16(vec3(lightmaps, encodeMaterialIDs(materialIDs))),
        encode3x16(specular),
        shadowed * 0.5 + 0.5
    );
}